import * as db from './db';
import { getSlackPostData } from './slack/service';
import { IncomingMessage } from 'http';
import { sendSuggestions, sendDiners, sendCityOptions, sendSuggestionsOptions } from './slack/bot';
import { suggestToUsers } from './service';
import { Auth, DbUser, User } from './types';
import { COMMANDS } from './constants';

export const stopCommand = async (request: IncomingMessage) => {
  const { verified, user_name, user_id, team_id } = await getSlackPostData(request, COMMANDS.stop);
  console.log(`command /stop by ${user_name}`);
  if ((await db.checkUserExists(user_id, team_id)) === db.ExistsType.NOT)
    return { text: 'Esmalt registreerige ennast kasutades käsklust /start' };
  const saved = verified && (await db.deleteUserRegister(user_id, team_id));
  return {
    text: saved ? 'Teie kasutaja deaktiveeriti' : 'Midagi läks valesti',
  };
};

export const startCommand = async (request: IncomingMessage) => {
  const { user_id, team_id, user_name, verified, text: postText } = await getSlackPostData(request, COMMANDS.start);
  console.log(`command /start ${postText} by ${user_name}`);

  if (!verified) {
    console.log('not verified');
    return { text: 'Midagi läks valesti!' };
  }

  const userData: DbUser = {
    user_id,
    team_id,
    user_name,
    city: null,
    district: null,
  };

  const userState = await db.checkUserExists(user_id, team_id);

  switch (userState) {
    case db.ExistsType.ACTIVE:
      const settings = await db.getUserSettings(user_id, team_id);
      if (!settings.city) {
        const bot_access_token = await db.getTeamBotToken(team_id);
        sendCityOptions(user_id, bot_access_token, settings.city);
        return true;
      }
      return { text: 'Soovitaja on juba lubatud.' };
    case db.ExistsType.NOT: {
      let ok = await db.saveUserRegister(userData);
      if (!ok) {
        console.log('save user failed');
        return { text: 'Midagi läks valesti' };
      }

      const bot_access_token = await db.getTeamBotToken(team_id);
      ok = await sendCityOptions(user_id, bot_access_token, null);

      return true;
    }
    default: {
      const saved = await db.unDeleteUserRegister(user_id, team_id);
      !saved && console.log('set user deleted_at failed');
      return { text: saved ? 'Soovitaja lubatud!' : 'Midagi läks valesti.' };
    }
  }
};

export const soovitaCommand = async (request: IncomingMessage) => {
  const { verified, team_id, user_id, user_name } = await getSlackPostData(request, COMMANDS.soovita);
  console.log(`command /soovita by ${user_name}`);
  if (!verified) return { text: 'Midagi läks valesti...' };
  if ((await db.checkUserExists(user_id, team_id)) === db.ExistsType.NOT)
    return { text: 'Esmalt registreerige ennast kasutades käsklust /start' };
  const bot_access_token = await db.getTeamBotToken(team_id);
  const auth: Auth = { bot_access_token, user_id, team_id, user_name };

  const settings = await db.getUserSettings(user_id, team_id);
  if (!settings.city) {
    sendCityOptions(user_id, bot_access_token, settings.city);
    return { text: 'Soovituste saamiseks pead valima linna' };
  }

  suggestToUsers([auth], settings.city);

  return true;
};

export const lemmikudCommand = async (request: IncomingMessage) => {
  const { verified, user_name, team_id, user_id } = await getSlackPostData(request, COMMANDS.lemmikud);
  console.log(`command /lemmikud by ${user_name}`);
  if (!verified) return { text: 'Midagi läks valesti...' };
  if ((await db.checkUserExists(user_id, team_id)) === db.ExistsType.NOT)
    return { text: 'Esmalt registreerige ennast kasutades käsklust /start' };

  const token = await db.getTeamBotToken(team_id);
  const user: User = { name: user_name, token };

  const favorites = await db.getUserFavorites(user_id, team_id);
  const diners = await db.getDinersById(favorites);
  const dinersWithFavorites = diners.map(diner => ({ ...diner, favorite: true }));
  let meals = await db.getTodaysMealsForDiners(diners.map(diner => diner.id));

  sendSuggestions(meals, dinersWithFavorites, user).then(console.log);

  return true;
};

export const otsiCommand = async (request: IncomingMessage) => {
  const { verified, user_name, team_id, text, user_id } = await getSlackPostData(request, COMMANDS.otsi);
  console.log(`command /otsi ${text} by ${user_name}`);
  if (!verified) return { text: 'Midagi läks valesti...' };
  if ((await db.checkUserExists(user_id, team_id)) === db.ExistsType.NOT)
    return { text: 'Esmalt registreerige ennast kasutades käsklust /start' };
  const input: string = text;

  const teamBotToken = await db.getTeamBotToken(team_id);
  const user: User = {
    name: user_name,
    token: teamBotToken,
  };
  const settings = await db.getUserSettings(user_id, team_id);
  const city = settings.city;
  if (!city) {
    sendCityOptions(user_id, teamBotToken, city);
    return { text: 'Toidukohtade otsimiseks pead valima linna' };
  }

  const diners = await db.findDiners(input, city);
  const meals = diners.length ? await db.getTodaysMealsForDiners(diners.map(diner => diner.id)) : [];
  const favorites = await db.getUserFavorites(user_id, team_id);
  const dinersWithFavorites = diners.map(diner => {
    return { ...diner, favorite: favorites.includes(parseInt(`${diner.id}`, 10)) };
  });

  sendDiners(dinersWithFavorites, meals, user).then(console.log);

  return true;
};

export const tahanCommand = async (request: IncomingMessage) => {
  const { verified, user_name, team_id, text, user_id } = await getSlackPostData(request, COMMANDS.tahan);
  console.log(`command /tahan ${text} by ${user_name}`);
  if (!verified) return { text: 'Midagi läks valesti...' };
  if ((await db.checkUserExists(user_id, team_id)) === db.ExistsType.NOT)
    return { text: 'Esmalt registreerige ennast kasutades käsklust /start' };
  const input: string = text.trim();
  if (!input) {
    return { text: 'Siseta mõni märksõna mida päevapakkumiste hulgast otsida. Näiteks /tahan salat' };
  }

  const teamBotToken = await db.getTeamBotToken(team_id);
  const user: User = {
    name: user_name,
    token: teamBotToken,
  };
  const settings = await db.getUserSettings(user_id, team_id);
  const city = settings.city;
  if (!city) {
    sendCityOptions(user_id, teamBotToken, city);
    return { text: 'Toitude otsimiseks pead valima linna' };
  }

  const meals = await db.findTodaysMeals(input, city);
  const diners = meals.length && (await db.getDinersById(meals.map(meal => meal.diner_id)));
  const favorites = await db.getUserFavorites(user_id, team_id);
  const dinersWithFavorites =
    diners &&
    diners.length &&
    diners.map(diner => {
      return { ...diner, favorite: favorites.includes(parseInt(`${diner.id}`, 10)) };
    });

  if (dinersWithFavorites) {
    sendDiners(dinersWithFavorites, meals, user).then(console.log);
  } else {
    return { text: 'Ei leidnud midagi...' };
  }

  return true;
};

export const seadedCommand = async (request: IncomingMessage) => {
  const { verified, user_name, team_id, user_id } = await getSlackPostData(request, COMMANDS.seaded);
  console.log(`command /seaded by ${user_name}`);
  if (!verified) return { text: 'Midagi läks valesti...' };
  if ((await db.checkUserExists(user_id, team_id)) === db.ExistsType.NOT)
    return { text: 'Esmalt registreerige ennast kasutades käsklust /start' };

  const teamBotToken = await db.getTeamBotToken(team_id);

  const settings = await db.getUserSettings(user_id, team_id);
  await sendCityOptions(user_id, teamBotToken, settings.city);
  await sendSuggestionsOptions(user_id, teamBotToken, settings.suggest_favorites);

  return true;
};
