require('dotenv').config();
import mysql, { MysqlError, FieldInfo } from 'mysql';
import {
  DbMeal,
  DbDiner,
  DbUser,
  DbChoice,
  OAuthResult,
  Auth,
  DbClassifiedMeal,
  DbClassifiedChoice,
  UserSettings,
  ClassifierResponse,
} from './types';

const db = mysql.createPool(process.env.DATABASE_URL || '');

function query(sql: string, values: any) {
  return new Promise<{ error: MysqlError | null; results?: any; fields?: FieldInfo[] }>(resolve => {
    db.query(sql, values, (error, results, fields) => {
      console.log(`db: ${sql.slice(0, 110)}`);
      if (error) {
        console.log({
          error: error ? { code: error.code, message: error.sqlMessage, sql: error.sql } : false,
          message: results && results.message ? results.message : '',
        });
      }
      resolve({ error, results, fields });
    });
  });
}

export async function saveOAuthResult(oauthResult: OAuthResult): Promise<boolean> {
  if (!oauthResult.ok) return false;
  const keys = ['team_id', 'access_token', 'scope', 'user_id', 'team_name', 'bot_user_id', 'bot_access_token'];
  const sql = `INSERT INTO oauth (${keys.join(', ')}) VALUES (${keys.map(_ => '?').join(', ')})`;
  const values = keys.map(key => oauthResult[key]);
  const { error } = await query(sql, values);
  return !error;
}

export async function deleteOAuthResult(teamId: string): Promise<boolean> {
  const sql = `DELETE FROM oauth WHERE team_id = ?`;
  const values = [teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function getUsersToSuggestTo(city: string): Promise<Array<Auth>> {
  const sql = `SELECT bot_access_token, users.user_id, users.team_id, users.user_name, users.suggest_favorites 
  FROM users LEFT JOIN oauth ON oauth.team_id = users.team_id 
  WHERE users.deleted_at IS NULL 
  AND users.suggested_at < CURDATE()
  AND users.city LIKE ?`;
  const { results } = await query(sql, [city]);
  return results?.map((o: any) => ({ ok: true, ...o })) || [];
}

export async function findDiners(name: string, city: string): Promise<Array<DbDiner>> {
  const sql = `SELECT * FROM diners WHERE name LIKE ? and city like ? LIMIT 5`;
  const { results } = await query(sql, [`%${name}%`, city]);
  return results || [];
}

export async function findTodaysMeals(name: string, city: string): Promise<Array<DbMeal>> {
  const sql = `SELECT max(meals.id) AS id, max(diner_id) as diner_id, max(price) as price, description 
  FROM meals 
  LEFT JOIN diners ON diners.id = meals.diner_id
  WHERE description LIKE ? 
  AND city like ?
  AND meals.created_at > CURDATE() 
  GROUP BY description
  LIMIT 10`;
  const { results } = await query(sql, [`%${name}%`, city]);
  return results || [];
}

export async function saveUserFavorite(userId: string, teamId: string, dinerId: Number): Promise<boolean> {
  const keys = ['diner_id', 'user_id', 'team_id'];
  const sql = `INSERT INTO favorites (${keys.join(', ')}) VALUES (${keys.map(_ => '?').join(', ')})`;
  const values = [dinerId, userId, teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function deleteUserFavorite(userId: string, teamId: string, dinerId: Number): Promise<boolean> {
  const sql = `DELETE FROM favorites WHERE diner_id = ? AND user_id = ? AND team_id = ?`;
  const values = [dinerId, userId, teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function getUserFavorites(userId: string, teamId: string): Promise<Array<number>> {
  const sql = `SELECT diner_id FROM favorites WHERE deleted_at IS NULL AND user_id = ? AND team_id = ?`;
  const values = [userId, teamId];
  const { results } = await query(sql, values);
  return results?.map((o: any) => o.diner_id) || [];
}

export enum ExistsType {
  NOT,
  ACTIVE,
  DELETED,
}

export async function checkUserExists(userId: string, teamId: string): Promise<ExistsType> {
  const sql = `SELECT user_id, team_id, deleted_at FROM users WHERE user_id = ? AND team_id = ?`;
  const values = [userId, teamId];
  const { results } = await query(sql, values);
  if (results?.length) {
    return results[0]['deleted_at'] !== null ? ExistsType.DELETED : ExistsType.ACTIVE;
  }
  return ExistsType.NOT;
}

export async function getTeamBotToken(teamId: string): Promise<string> {
  const column = 'bot_access_token';
  const sql = `SELECT ${column} FROM oauth WHERE team_id = ?`;
  const values = [teamId];
  const { results } = await query(sql, values);
  if (results?.length) {
    return results[0][column];
  }
  return '';
}

export async function getUserSettings(userId: string, teamId: string): Promise<UserSettings> {
  const sql = `SELECT city, suggest_favorites FROM users WHERE user_id = ? AND team_id = ?`;
  const values = [userId, teamId];
  const { results } = await query(sql, values);
  return results[0];
}

export async function saveUserChoice(data: DbChoice): Promise<boolean> {
  const keys = ['meal_id', 'user_id', 'team_id', 'suggestions'];
  const sql = `INSERT INTO choices (${keys.join(', ')}) VALUES (${keys.map(_ => '?').join(', ')})`;
  const values = keys.map(key => data[key]);
  const { error } = await query(sql, values);
  return !error;
}

export async function refreshUserSuggestedAt(userId: string, teamId: string): Promise<boolean> {
  const sql = `UPDATE users SET suggested_at = CURRENT_TIMESTAMP WHERE user_id = ? AND team_id = ?`;
  const values = [userId, teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function deleteUserRegister(userId: string, teamId: string): Promise<boolean> {
  const sql = `UPDATE users SET deleted_at = CURRENT_TIMESTAMP WHERE user_id = ? AND team_id = ?`;
  const values = [userId, teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function unDeleteUserRegister(userId: string, teamId: string): Promise<boolean> {
  const sql = `UPDATE users SET deleted_at = null WHERE user_id = ? AND team_id = ?`;
  const values = [userId, teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function updateUserCity(userId: string, teamId: string, city: string): Promise<boolean> {
  const sql = `UPDATE users SET city = ? WHERE user_id = ? AND team_id = ?`;
  const values = [city, userId, teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function updateUserSuggestFavorites(
  userId: string,
  teamId: string,
  suggest_favorites: boolean
): Promise<boolean> {
  const sql = `UPDATE users SET suggest_favorites = ? WHERE user_id = ? AND team_id = ?`;
  const values = [suggest_favorites, userId, teamId];
  const { error } = await query(sql, values);
  return !error;
}

export async function saveUserRegister(data: DbUser): Promise<boolean> {
  const keys = ['user_id', 'team_id', 'user_name', 'city', 'district'];
  const sql = `INSERT INTO users (${keys.join(', ')}) VALUES (${keys.map(_ => '?').join(', ')})`;
  const values = keys.map(key => data[key]);
  const { error } = await query(sql, values);
  return !error;
}

export async function getAllDiners(city: string): Promise<Array<DbDiner>> {
  const sql = `SELECT * FROM diners WHERE city LIKE ?`;
  const { results } = await query(sql, [city]);
  return results || [];
}

export async function saveDiners(diners: Array<DbDiner>): Promise<boolean> {
  const keys = ['id', 'name', 'address', 'district', 'homepage', 'h_coordinate', 'l_coordinate', 'city', 'name_as_url'];
  const sql = `INSERT INTO diners (${keys.join(', ')}) VALUES ?`;
  const values = diners.map(diner => keys.map(key => diner[key]));
  const { error } = await query(sql, [values]);
  return !error;
}

export async function saveMeals(meals: Array<DbMeal>): Promise<boolean> {
  const keys = ['diner_id', 'description', 'price'];
  const sql = `INSERT INTO meals (${keys.join(', ')}) VALUES ?`;
  const values = meals.map(meal => keys.map(key => meal[key]));
  const { error } = await query(sql, [values]);
  return !error;
}

export async function getUnclassifiedMeals(): Promise<Array<{ id: number; description: string }>> {
  const sql = `SELECT m.id as id, m.description as description 
    FROM meals m 
    LEFT JOIN classifications c ON c.meal_id = m.id 
    WHERE c.id IS NULL
    AND m.created_at > CURDATE()`;
  const { results } = await query(sql, []);
  return results;
}

export async function saveClassifications(classification: ClassifierResponse): Promise<boolean> {
  const keys = ['meal_id', 'classifier', 'category', 'lemmas', 'brackets', 'meat', 'ingredients', 'vegetarian'];
  const sql = `INSERT IGNORE INTO classifications (${keys.join(', ')}) VALUES ?`;
  const values = classification.meals.map(meal => [
    meal.id,
    meal.classifier,
    meal.category,
    JSON.stringify(meal.lemmas),
    JSON.stringify(meal.brackets),
    JSON.stringify(meal.meat),
    JSON.stringify(meal.ingredients),
    meal.vegetarian,
  ]);
  const { error } = await query(sql, [values]);
  return !error;
}

export async function getTodaysMeals(): Promise<Array<DbMeal>> {
  const sql = `SELECT * FROM meals WHERE created_at > CURDATE()`;
  const { results } = await query(sql, []);
  return results;
}

export async function getUsersChoicesClassified(
  userIds: Array<string>,
  days: number = 10
): Promise<Array<DbClassifiedChoice>> {
  const sql = `SELECT choices.user_id, choices.team_id, meals.id, diner_id, price, description, classifier, category, lemmas, 
  brackets, meat, ingredients, vegetarian, l_coordinate, h_coordinate, city, 
  district, suggestions, choices.created_at

  FROM choices
  LEFT JOIN meals on choices.meal_id = meals.id
  LEFT JOIN classifications c ON c.meal_id = meals.id 
  LEFT JOIN diners d ON d.id = meals.diner_id
  WHERE c.id IS NOT NULL
  AND choices.user_id in ?
  AND choices.created_at > CURDATE() - INTERVAL ? DAY
  ORDER BY choices.id DESC
  `;
  const { results } = await query(sql, [[userIds], days]);
  return results.map(
    (result: any): DbClassifiedChoice => ({
      user_id: result.user_id,
      team_id: result.team_id,
      id: result.id,
      diner_id: result.diner_id,
      price: result.price,
      description: result.description,
      classifier: result.classifier,
      category: result.category,
      lemmas: JSON.parse(result.lemmas),
      brackets: JSON.parse(result.brackets),
      meat: JSON.parse(result.meat),
      ingredients: JSON.parse(result.ingredients),
      vegetarian: result.vegetarian,
      l_coordinate: result.l_coordinate ? Number(result.l_coordinate) : null,
      h_coordinate: result.h_coordinate ? Number(result.h_coordinate) : null,
      suggestions: JSON.parse(result.suggestions),
      created_at: result.created_at,
    })
  );
}

export async function getTodaysMealsWithClassifications(city: string): Promise<Array<DbClassifiedMeal>> {
  const sql = `SELECT meals.id, diner_id, price, description, classifier, category, lemmas, brackets, meat, ingredients, vegetarian, l_coordinate, h_coordinate, city, district
  FROM meals 
  LEFT JOIN classifications c ON c.meal_id = meals.id 
  LEFT JOIN diners d ON d.id = meals.diner_id
  WHERE c.id IS NOT NULL 
  AND meals.created_at > CURDATE()
  AND d.city LIKE ?
  `;
  const { results } = await query(sql, [city]);
  return results.map((result: any) => ({
    id: result.id,
    diner_id: result.diner_id,
    price: result.price,
    description: result.description,
    classifier: result.classifier,
    category: result.category,
    lemmas: JSON.parse(result.lemmas),
    brackets: JSON.parse(result.brackets),
    meat: JSON.parse(result.meat),
    ingredients: JSON.parse(result.ingredients),
    vegetarian: result.vegetarian,
    l_coordinate: result.l_coordinate ? Number(result.l_coordinate) : null,
    h_coordinate: result.h_coordinate ? Number(result.h_coordinate) : null,
  }));
}

export async function getTodaysMealsForDiners(ids: Array<Number>): Promise<Array<DbMeal>> {
  const sql = `SELECT * FROM meals WHERE created_at > CURDATE() AND diner_id IN ?`;
  const { results } = await query(sql, [[ids]]);
  return results || [];
}

export async function getDinersById(ids: Array<number>): Promise<Array<DbDiner>> {
  const sql = `SELECT * FROM diners WHERE id IN ?`;
  const { results } = await query(sql, [[ids]]);
  return results || [];
}
