import { IncomingMessage } from 'http';
import querystring from 'querystring';
import { StringDecoder } from 'string_decoder';

export class Timer {
  logging = false;
  timers: { [index: string]: number } = {};
  keys: Array<string> = [];
  timerName = '';

  setLogging(value: boolean) {
    this.logging = value;
    return this;
  }

  setName(name: string = '') {
    this.timerName = name;
    return this;
  }

  start() {
    this.timers = {};
    this.keys = [];
    this.timers['start'] = Date.now();
    this.keys.push('start');
    return this;
  }

  checkpoint(name: string) {
    if (!this.keys.length) throw new Error('Timer not started');
    if (this.keys.includes(name)) throw new Error(`Already have ${name}`);
    this.timers[name] = Date.now();
    this.keys.push(name);
    this._logLatest();
    return this;
  }

  _logLatest() {
    if (this.logging) {
      return this.log();
    }
    return this;
  }

  log() {
    const prevKey = this.keys[this.keys.length - 2];
    const lastKey = this.keys[this.keys.length - 1];
    const time = this.timers[lastKey] - this.timers[prevKey];
    console.log(`${this.timerName ? `${this.timerName}: ` : ''}${lastKey} ${time}ms`);
    return this;
  }

  results() {
    if (!this.keys.length) throw new Error('Timer not started');
    this.timers['end'] = Date.now();
    this.keys.push('end');
    const times: { [index: string]: number } = {};
    for (let i = 1; i < this.keys.length - 1; i++) {
      times[this.keys[i]] = this.timers[this.keys[i]] - this.timers[this.keys[i - 1]];
    }

    times.total = this.timers['end'] - this.timers['start'];

    const result = { timers: { ...this.timers }, times, total: times.total, name: this.timerName };
    this.timers = {};
    this.keys = [];

    return result;
  }

  getCurrent() {
    return Date.now() - this.timers['start'];
  }
}

export async function getPostData(request: IncomingMessage): Promise<any> {
  return new Promise(resolve => {
    let buffer = '';
    const decoder = new StringDecoder('utf-8');

    if (request.method === 'POST') {
      request.on('data', function(chunk) {
        buffer += decoder.write(chunk);
        if (buffer.length > 1e6) {
          buffer = '';
          request.connection.destroy();
        }
      });

      request.on('end', function() {
        try {
          buffer += decoder.end();
          const contentType = request.headers['content-type'];
          if (contentType === 'application/json') resolve(JSON.parse(buffer));
          else if (contentType === 'application/x-www-form-urlencoded') resolve(querystring.parse(buffer));
          else resolve({ error: `can not parse content-type: ${contentType}` });
        } catch (err) {
          console.log(buffer);
          resolve({ error: true });
        }
      });
    } else {
      resolve({});
    }
  });
}

export function getParams(url?: string): any {
  if (!url) return {};
  const params: any = {};
  const rawParams = url.split('?')[1] || '';
  rawParams.split('&').forEach(param => {
    const [key, value] = param.split('=');
    params[key] = value;
  });
  return params;
}
