import axios from 'axios';
import * as db from './db';
import {
  Auth,
  DbClassifiedMeal,
  DbClassifiedChoice,
  Suggestions,
  RecommenderResponse,
  RecommenderRequest,
  ClassifierResponse,
  ClassifierRequest,
} from './types';

const url = 'http://pp-classifier.xn--lll-qla.ee:5000/';
const instance = axios.create({
  baseURL: url,
  timeout: 10000,
  validateStatus: _status => {
    return true;
  },
  headers: { Accept: 'application/json' },
});

function chunkArray<T>(arr: Array<T>, n: number): Array<Array<T>> {
  const chunks = [];
  for (let i = 0; i < arr.length; i += n) {
    chunks.push(arr.slice(i, i + n));
  }
  return chunks;
}

export async function classify(data: Array<{ id: number; description: string }>) {
  const chunks = chunkArray(data, 100);
  process.stdout.write(`classifying ${data.length} meals in ${chunks.length} chunks`);

  const results = await Promise.all(
    chunks.map(async chunk => {
      const request: ClassifierRequest = {
        meals: chunk.map(meal => ({
          id: meal.id,
          description: meal.description,
        })),
      };
      const response = await instance.post<ClassifierResponse>('classify', request);

      process.stdout.write(` (${response.status})\n`);

      if (response.status === 200) {
        const result = response.data;
        const errors = result.meals.filter(meal => meal.error);
        const filtered = result.meals.filter(meal => !meal.error);
        if (errors.length) {
          console.log('Errors in classification!');
          console.log(JSON.stringify(errors));
        }
        if (filtered.length > 0) return await db.saveClassifications({ meals: filtered });
        return true;
      } else {
        console.error(response.data, response.statusText);
        throw new Error('Failed saving results');
      }
    })
  );
  return results;
}

export function pick(n: number, arr: Array<any>): Array<any> {
  const nr = n > arr.length ? arr.length : n;
  const indexes = arr.map((_, i) => i);
  const picked = [];
  for (let i = 0; i < nr; i++) picked.push(indexes.splice(randInt(0, indexes.length - 1), 1)[0]);
  return picked.map(x => arr[x]);
}

export function randInt(min: number, max: number): number {
  let n1: number, n2: number;
  if (min > max) {
    n1 = max;
    n2 = min + 1;
  } else {
    n1 = min;
    n2 = max + 1;
  }
  const diff = n2 - n1;
  return n1 + Math.floor(Math.random() * diff);
}

export async function getRecommendations(
  users: Auth[],
  meals: DbClassifiedMeal[],
  choices: DbClassifiedChoice[],
  debug: boolean
): Promise<Array<Suggestions>> {
  process.stdout.write(
    `running recommender for ${users.length} users with ${meals.length} meals and ${choices.length} choices... `
  );

  const payload: RecommenderRequest = { users, meals, choices, debug };
  const response = await instance.post<RecommenderResponse>('recommend', payload);

  process.stdout.write(`${response.status}`);

  if (response.status === 200) {
    console.log(response.data?.errors);
    if (debug) {
      response.data.suggestions.forEach(sug => {
        console.log(JSON.stringify(sug));
      });
    }
    return response.data.suggestions;
  }

  return [];
}
