require('dotenv').config();
import { createServer, IncomingMessage, ServerResponse } from 'http';
import { getParams } from './functions';
import * as controller from './controller';
import * as commands from './commands';
import { slackRequestListener } from './slack/requestListener';
import { initSuggestQueue } from './queue';
import { COMMANDS } from './constants';

type Route = (request: IncomingMessage, response: ServerResponse) => any;
type Routes = { [index: string]: Route | undefined };
type Methods = { [index in 'GET' | 'POST']: Routes };

const methods: Methods = {
  GET: {
    '/auth': controller.authorize,
    '/': controller.getSlackLink,
  },
  POST: {
    '/interact': slackRequestListener,
    [COMMANDS.start]: commands.startCommand,
    [COMMANDS.stop]: commands.stopCommand,
    [COMMANDS.soovita]: commands.soovitaCommand,
    [COMMANDS.lemmikud]: commands.lemmikudCommand,
    [COMMANDS.otsi]: commands.otsiCommand,
    [COMMANDS.tahan]: commands.tahanCommand,
    [COMMANDS.seaded]: commands.seadedCommand,
  },
};

const createListener = (methods: Methods) => async (request: IncomingMessage, response: ServerResponse) => {
  const { method, url }: { method: 'GET' | 'POST'; url: string } = request as any;
  const route = url.split('?')[0];
  const routes = methods[method];
  const _function = routes[route];
  try {
    const result =
      typeof _function === 'function'
        ? await _function(request, response)
        : { url, route, params: getParams(url), method, message: 'route not found' };

    if (typeof result === 'object') {
      response.writeHead(200, { 'Content-Type': 'application/json' }).end(JSON.stringify(result), console.log);
    } else if (typeof result === 'boolean') {
      response.writeHead(200).end();
    }
  } catch (e) {
    console.error(e);
    response.writeHead(500, { 'Content-Type': 'application/json' }).end(JSON.stringify({ message: e.message }));
  }
};

const PORT = process.env.PORT || 8080;

const requestListener = createListener(methods);
const server = createServer(requestListener);

server.listen(PORT, () => {
  console.log('Server listening on: http://localhost:%s', PORT);
  initSuggestQueue().then(() => console.log('Queue initialized'));
});
