import axios from 'axios';
import * as db from './db';
import { MealsResponse, DbDiner, DbMeal, Auth } from './types';
import { getRecommendations, classify } from './recommender';
import { sendSuggestions } from './slack/bot';
import { Timer } from './functions';
import { CITIES } from './constants';

const url = process.env.PP_API_URL;
const token = process.env.PP_API_TOKEN;

const instance = axios.create({
  baseURL: url,
  timeout: 10000,
  headers: {
    'X-Authorization': token,
    Accept: 'application/json',
  },
});

export const suggestToUsers = async (
  users: Array<Auth>,
  city: string
): Promise<{ ok: number; failed: number; times: any }> => {
  const logging = users.length === 1;
  let timer = new Timer()
    .setName(`suggestToUsers(${city})`)
    .setLogging(logging)
    .start();

  const meals = await db.getTodaysMealsWithClassifications(city);
  timer.checkpoint(`Loaded todays meals for ${city}`);
  if (!meals.length) return { ok: 0, failed: 0, times: timer.results() };

  const userIds = users.map(user => user.user_id);
  const diners = await db.getAllDiners(city);
  timer.checkpoint(`Loaded all diners in ${city}`);
  const usersChoices = await db.getUsersChoicesClassified(userIds, 100);
  timer.checkpoint(`Loaded 100 days choices for ${userIds.length} users`);

  const allSuggestions = await getRecommendations(users, meals, usersChoices, logging);
  timer.checkpoint('Received recommendations response');

  const sendingPromises = users.map(async auth => {
    const sendingTimer = new Timer()
      .setLogging(logging)
      .setName(`- Serving ${auth.user_name}`)
      .start();
    const favorites = await db.getUserFavorites(auth.user_id, auth.team_id);
    sendingTimer.checkpoint('Loaded user favorites');
    const recommenderSuggestions = allSuggestions.find(
      suggestion => suggestion.user_id === auth.user_id && suggestion.team_id === auth.team_id
    );
    sendingTimer.checkpoint('Found users suggestions');

    let suggestions: Array<DbMeal> = [];
    if (favorites.length && auth.suggest_favorites) {
      logging && console.log(`${auth.user_name} has favorites and want to receive favorites`);
      suggestions = meals.filter(meal => favorites.includes(meal.diner_id));
      sendingTimer.checkpoint('Filtered all meals by favorites');
    } else {
      if (recommenderSuggestions?.meal_ids?.length) {
        const mealIds = recommenderSuggestions.meal_ids.slice(0, 10);
        suggestions = meals.filter(meal => mealIds.includes(meal.id));
        sendingTimer.checkpoint('Filtered by recommendations');
        if (auth.suggest_favorites) {
          console.log(`${auth.user_name} wanted favorites but has no favorites selected.`);
        } else {
          console.log(`${auth.user_name} wanted suggestions and gottem.`);
        }
      } else {
        console.log(`${auth.user_name} had no recommendations!`);
      }
    }

    const relevantDiners = diners.filter(diner => suggestions.map(meal => meal.diner_id).includes(diner.id));
    const dinersWithFavoritesMarked = relevantDiners.map(diner => ({
      ...diner,
      favorite: favorites.includes(diner.id),
    }));

    let ok = false;
    try {
      ok = await sendSuggestions(suggestions, dinersWithFavoritesMarked, {
        name: auth.user_id,
        token: auth.bot_access_token,
      });
      ok = ok && (await db.refreshUserSuggestedAt(auth.user_id, auth.team_id));
    } catch (e) {
      console.error(e);
    }

    sendingTimer.checkpoint(`${ok ? 'Sent' : 'Error sending '} suggestions`);
    const { total } = sendingTimer.results();

    console.log({ user_name: auth.user_name, user_id: auth.user_id, team_id: auth.team_id, ok, time: `${total}ms` });
    return ok;
  });

  const results = await Promise.all(sendingPromises);
  timer.checkpoint('Sending done');
  const ok = results.reduce((previous, current) => previous + (current ? 1 : 0), 0);
  const failed = results.reduce((previous, current) => previous + (current ? 0 : 1), 0);
  return { ok, failed, times: timer.results() };
};

export const updateMeals = async (city: string) => {
  const mealsResponse = await getAllMealsInCity(city);
  const savedDiners = await saveDiners(mealsResponse, city);
  const savedMeals = savedDiners && (await saveMeals(mealsResponse));
  const unclassifiedMeals = await db.getUnclassifiedMeals();
  const classifiedMeals = savedMeals && (await classify(unclassifiedMeals));
  const saved = !!(savedDiners && savedMeals && classifiedMeals);
  console.log(
    saved
      ? 'OK'
      : `Failed at ${[savedDiners && 'diners', savedMeals && 'meals', !!classifiedMeals && 'classify']
          .filter(x => x)
          .join(', ')}`
  );
  return { saved };
};

const suggestToAllUsers = async (city: string) => {
  console.log(`Getting users to suggest to in ${city}`);
  const users = await db.getUsersToSuggestTo(city);
  console.log(`Found ${users.length} users`);
  if (!users.length) return { ok: 0, failed: 0, times: {} };
  const { ok, failed, times } = await suggestToUsers(users, city);
  return { ok, failed, times };
};

export const suggestToAllCities = async (progress?: (value: number) => void) => {
  console.log('Suggest to all cities');
  let i = 0;
  let result;
  let results = [];
  for (const city of CITIES) {
    console.log(`- Suggest to ${city}`);
    result = await suggestToAllUsers(city);
    results.push(result);
    i += 1;
    if (result.ok) await progress?.((i / CITIES.length) * 100);
  }

  return results;
};

export async function getAllMealsInCity(city: string): Promise<Array<MealsResponse>> {
  process.stdout.write('Fetching meals from pp api ');
  const mealsRaw = await instance.get<{ offerings: Array<MealsResponse> }>(`andmed/slack/linnad/${city}`);
  process.stdout.write(`(${mealsRaw.status})\n`);
  const data: Array<MealsResponse> = mealsRaw.data.offerings;
  return data;
}

export async function saveDiners(mealsResponse: Array<MealsResponse>, city: string): Promise<boolean> {
  const existingDiners = await db.getAllDiners('%');
  const existingDinerIds = existingDiners.map(diner => diner.id);
  const unsavedDiners = mealsResponse
    .filter(response => !existingDinerIds.includes(response.diner.id))
    .map(
      (response): DbDiner => ({
        id: response.diner.id,
        name: response.diner.name,
        address: response.diner.address || '',
        district: response.diner.district || '',
        homepage: response.diner.homepage || '',
        h_coordinate: response.diner.coordinates?.h || null,
        l_coordinate: response.diner.coordinates?.l || null,
        city: city || '',
        name_as_url: response.diner.nameAsUrl || '',
      })
    );

  return unsavedDiners.length > 0 ? db.saveDiners(unsavedDiners) : true;
}

export async function saveMeals(mealsResponse: Array<MealsResponse>): Promise<boolean> {
  const existingMeals = await db.getTodaysMeals();
  const meals: Array<DbMeal> = [];
  mealsResponse.forEach(response => {
    response.meals.forEach(mealResponse => {
      const meal: DbMeal = {
        diner_id: response.diner.id,
        price: mealResponse.price || '',
        description: mealResponse.description || '',
      };
      const mealExists =
        -1 !==
        existingMeals.findIndex(
          existingMeal =>
            meal.diner_id === existingMeal.diner_id &&
            meal.description === existingMeal.description &&
            meal.price === existingMeal.price
        );

      if (!mealExists) meals.push(meal);
    });
  });

  return meals.length > 0 ? db.saveMeals(meals) : true;
}
