import { createMessageAdapter } from '@slack/interactive-messages';
import * as db from '../db';
import { CITIES, getCityName } from '../constants';

const slackSigningSecret = process.env.SLACK_SIGNING_SECRET || '';
const slackInteractions = createMessageAdapter(slackSigningSecret);
const regexp: RegExp = /.*/;

slackInteractions.action(regexp, async (payload, respond) => {
  const { token, user, actions, message } = payload;
  const verified = token === process.env.SLACK_VERIFICATION_TOKEN;

  if (verified) {
    try {
      const actionObj = actions[0];
      const type = actionObj.type;
      console.log(`Slack request listener handling ${type}`);
      const value = type === 'button' ? actionObj.value : actionObj.selected_option.value;

      const [action, ...actionPayload] = value.split('|');
      const blocks: Array<any> = message.blocks;
      const actionBlockObject = blocks.find(block => block.block_id === actionObj.block_id);

      switch (action) {
        case 'dismiss': {
          await respond({ delete_original: true });
          break;
        }
        case 'suggest_favorites': {
          const [suggest_favorites] = actionPayload;
          let ok;
          if (!['0', '1'].includes(suggest_favorites)) {
            ok = false;
          } else {
            ok = await db.updateUserSuggestFavorites(user.id, user.team_id, suggest_favorites === '1');
          }

          await respond({
            text: ok
              ? `Lõuna ajal saad edaspidi pakkumisi ${suggest_favorites === '1' ? 'lemmikutest' : 'soovitajalt'}.`
              : 'Midagi läks valesti...',
            replace_original: true,
          });
          break;
        }
        case 'city': {
          const [city] = actionPayload;
          let ok;
          if (!CITIES.includes(city)) {
            ok = false;
          } else {
            ok = await db.updateUserCity(user.id, user.team_id, city);
          }

          await respond({
            text: ok ? `Valisid ${getCityName(city)}` : 'Midagi läks valesti...',
            replace_original: true,
          });
          break;
        }
        case 'choose': {
          const [meal_id, suggestionsRaw] = actionPayload;
          const saved = await db.saveUserChoice({
            team_id: user.team_id,
            user_id: user.id,
            meal_id: meal_id === 'undefined' ? null : meal_id,
            suggestions: JSON.stringify(
              suggestionsRaw
                .split(',')
                .filter((value: string) => value !== '')
                .map((value: string) => parseInt(value, 10))
            ),
          });
          if (saved) {
            const removeAccessories = blocks.filter(
              block => !!block.accessory && block.accessory.type === 'button' && block.accessory.text?.text === 'Vali'
            );
            actionBlockObject.text.text = `\`${actionBlockObject.text.text}\``;
            for (let block of removeAccessories) delete block.accessory;
            await respond({ text: 'Naudi oma lõunat!', blocks, replace_original: true });
          } else await respond({ text: 'Midagi läks valesti...', replace_original: true });
          break;
        }
        case 'favorite': {
          const [diner_id] = actionPayload;
          const saved = await db.saveUserFavorite(user.id, user.team_id, diner_id);
          if (saved) {
            actionBlockObject.accessory.text.text = 'Eemalda lemmikutest';
            actionBlockObject.accessory.value = `unfavorite|${diner_id}`;
            await respond({ text: 'Lemmikutesse lisatud!', blocks, replace_original: true });
          } else await respond({ text: 'Midagi läks valesti...', replace_original: true });
          break;
        }
        case 'unfavorite': {
          const [diner_id] = actionPayload;
          const saved = await db.deleteUserFavorite(user.id, user.team_id, diner_id);
          if (saved) {
            actionBlockObject.accessory.text.text = 'Lisa lemmikutesse';
            actionBlockObject.accessory.value = `favorite|${diner_id}`;
            await respond({ text: 'Lemmikutest eemaldatud!', blocks, replace_original: true });
          } else await respond({ text: 'Midagi läks valesti...', replace_original: true });
          break;
        }
        default:
          await respond({ delete_original: true });
      }
    } catch (e) {
      await respond({ text: 'Midagi läks väga valesti...', replace_original: true });
      console.error(e);
      throw e;
    }
  } else {
    respond({ text: 'Action verification failed.', replace_original: true });
  }
});

export const slackRequestListener = slackInteractions.requestListener();
