from nltk import FreqDist
import operator
from math import pi, sin, cos, sqrt, atan2

class Recommender:
  name = 'DumbRecommender'

  def __init__(self, users, meals, choices, debug):
    self.users = users
    self.meals = meals
    self.choices = choices
    self.debug = not not debug

  def __getFeatures(self, user_id, team_id):
    user_choices = list(filter(lambda choice: choice['user_id'] == user_id and choice['team_id'] == team_id, self.choices))

    if len(user_choices) == 0:
      return None

    lemmas = []
    categories = []
    l_coordinates = []
    h_coordinates = []
    max_distance = 0
    min_distance = 1000000 # 1000km
    avg_distance = 0
    vegan = 0
    vegan_maybe = 0
    meat = []
    meat_count = 0

    for choice in user_choices:
      lemmas += list(set(choice['lemmas']))
      meat += choice['meat']
      meat_count += 1 if len(meat) else 0
      categories.append(choice['category'])

      has_coordinates = choice['l_coordinate'] is not None and choice['h_coordinate'] is not None
      if has_coordinates:
        l_coordinates.append(choice['l_coordinate'])
        h_coordinates.append(choice['h_coordinate'])

      if choice['vegetarian']:
        vegan += 1
      if choice['vegetarian'] is None:
        vegan_maybe += 1
    
    l_coordinate = sum(l_coordinates) / max(len(l_coordinates), 1)
    h_coordinate = sum(h_coordinates) / max(len(h_coordinates), 1)

    for choice in user_choices:
      has_coordinates = choice['l_coordinate'] is not None and choice['h_coordinate'] is not None
      if not has_coordinates:
        continue
      distance = self.__getCoordinatesDistanceInMeters(l_coordinate, h_coordinate, choice['l_coordinate'], choice['h_coordinate'])
      max_distance = max(distance, max_distance)
      min_distance = min(distance, min_distance)
      avg_distance += distance
    avg_distance /= len(user_choices)
    
    return {
      "lemma_freq": FreqDist(lemmas),
      "category_freq": FreqDist(categories),
      "l_coordinate": l_coordinate,
      "h_coordinate": h_coordinate,
      "max_distance": max_distance,
      "min_distance": min_distance,
      "avg_distance": avg_distance,
      "choices_count": len(user_choices),
      "vegan_count": vegan,
      "vegan_maybe_count": vegan_maybe,
      "meat_count": meat_count,
      "meat_freq" : FreqDist(meat)
    }
  
  def __getCoordinatesDistanceInMeters(self, lat1, lon1, lat2, lon2):
    R = 6371e3 # metres
    fi1 = (lat1 / 180) * pi
    fi2 = (lat2 / 180) * pi
    deltaFi = ((lat2 - lat1) / 180) * pi
    deltaLambda = ((lon2 - lon1) / 180) * pi
    a = sin(deltaFi / 2) * sin(deltaFi / 2) + cos(fi1) * cos(fi2) * sin(deltaLambda / 2) * sin(deltaLambda / 2)
    c = 2 * atan2(sqrt(a), sqrt(1 - a))
    d = R * c
    return d # meters

  def __getScore(self, features, meal):
    if features is None:
      return {
        "id": meal['id'],
        "score": 0
      }
    
    lemma_score = 0
    lemmas = set(meal['lemmas'])
    for lemma in lemmas:
      lemma_score += features['lemma_freq'][lemma]
    lemma_score /= len(lemmas) + len(features['lemma_freq'])
    
    # max 1 point
    category_score = features['category_freq'][meal['category']] / features['choices_count']

    # distance score calculation
    has_coordinates = meal['l_coordinate'] is not None and meal['h_coordinate'] is not None and features['max_distance'] != 0
    if has_coordinates:
      distance = self.__getCoordinatesDistanceInMeters(
        features['l_coordinate'], features['h_coordinate'],
        meal['l_coordinate'], meal['h_coordinate'])
      distance_score = (features['avg_distance'] - distance) / features['max_distance']
    else:
      distance_score = 0

    if len(meal['meat']) > 0:
      meat_prob = 1
    elif meal['vegetarian'] is False:
      meat_prob = 0.75
    elif meal['vegetarian'] is None:
      meat_prob = 0.5
    else:
      meat_prob = 0

    veg_percent = (features['vegan_count'] + features['vegan_maybe_count']) / features['choices_count']
    meat_score = 4 * (meat_prob - 0.5) * (0.5 - veg_percent)

    score = lemma_score + category_score + distance_score + meat_score
    return {
      "id": meal['id'],
      "lemma_score": lemma_score,
      "category_score": category_score,
      "distance_score": distance_score,
      "meat_score": meat_score,
      "score": score
    }

  def getSuggestions(self, user):
    user_id = user['user_id']
    team_id = user['team_id']

    user_features = self.__getFeatures(user_id, team_id)

    scores = []
    for meal in self.meals:
      scores.append(self.__getScore(user_features, meal))

    scores.sort(key = lambda x: x["score"], reverse=True)
    top10 = scores[0:10]
    
    result = {
      "user_id": user_id,
      "team_id": team_id,
      "meal_ids": [score['id'] for score in top10],
    }

    if self.debug:
      result['debug'] = { 
        "user_features": user_features,
        "meal_scores": top10
      }
    
    return result